# -*- coding: utf-8 -*-
from builtins import object

from ..utils import get_matrix_in_format
from sklearn.base import BaseEstimator


class GraphBuilderBase(object):
    """An abstract base class for a graph building class used in Label Space clustering.

    Inherit it in your classifier according to`developer guide <../developer.ipynb>`_.

    """

    def __init__(self):
        super(GraphBuilderBase, self).__init__()

    def transform(self, y):
        """Abstract method for graph edge map builder for a label space clusterer

        Implement it in your classifier according to`developer guide <../developer.ipynb>`_.

        Raises
        ------
        NotImplementedError
            this is an abstract method
        """
        raise NotImplementedError("GraphBuilderBase::transform()")


class LabelSpaceClustererBase(BaseEstimator):
    """An abstract base class for Label Space clustering

    Inherit it in your classifier according to`developer guide <../developer.ipynb>`_.

    """

    def __init__(self):
        super(LabelSpaceClustererBase, self).__init__()

    def fit_predict(self, X, y):
        """Abstract method for clustering label space

        Implement it in your classifier according to`developer guide <../developer.ipynb>`_.

        Raises
        ------
        NotImplementedError
            this is an abstract method
        """
        raise NotImplementedError("LabelSpaceClustererBase::fit_predict()")


class LabelGraphClustererBase(object):
    """An abstract base class for Label Graph clustering

    Inherit it in your classifier according to`developer guide <../developer.ipynb>`_.

    """

    def __init__(self, graph_builder):
        """

        Attributes
        ----------
        graph_builder : a  GraphBuilderBase derivative class
            a graph building class for the clusterer
        """
        super(LabelGraphClustererBase, self).__init__()
        self.graph_builder = graph_builder

    def fit_predict(self, X, y):
        """Abstract method for clustering label space

        Implement it in your classifier according to`developer guide <../developer.ipynb>`_.

        Raises
        ------
        NotImplementedError
            this is an abstract method
        """
        raise NotImplementedError("LabelGraphClustererBase::fit_predict()")


class LabelCooccurrenceGraphBuilder(GraphBuilderBase):
    """Base class providing API and common functions for all label
    co-occurence based multi-label classifiers.

    This graph builder constructs a Label Graph based on the output matrix where two label nodes are connected
    when at least one sample is labeled with both of them. If the graph is weighted, the weight of an edge between two
    label nodes is the number of samples labeled with these two labels. Self-edge weights contain the number of samples
    with a given label.

    Parameters
    ----------
    weighted: bool
        decide whether to generate a weighted or unweighted graph.
    include_self_edges : bool
        decide whether to include self-edge i.e. label 1 - label 1 in
        co-occurrence graph
    normalize_self_edges: bool
        if including self edges, divide the (i, i) edge by 2.0, requires include_self_edges=True


    References
    ----------

    If you use this graph builder please cite the clustering paper:

    .. code:: latex

        @Article{datadriven,
            author = {Szymański, Piotr and Kajdanowicz, Tomasz and Kersting, Kristian},
            title = {How Is a Data-Driven Approach Better than Random Choice in
            Label Space Division for Multi-Label Classification?},
            journal = {Entropy},
            volume = {18},
            year = {2016},
            number = {8},
            article_number = {282},
            url = {http://www.mdpi.com/1099-4300/18/8/282},
            issn = {1099-4300},
            doi = {10.3390/e18080282}
        }


    Examples
    --------

    A full example of building a modularity-based label space division based on the Label Co-occurrence Graph and
    classifying with a separate classifier chain per subspace.

    .. code :: python

        from skmultilearn.cluster import LabelCooccurrenceGraphBuilder, NetworkXLabelGraphClusterer
        from skmultilearn.ensemble import LabelSpacePartitioningClassifier
        from skmultilearn.problem_transform import ClassifierChain
        from sklearn.naive_bayes import GaussianNB

        graph_builder = LabelCooccurrenceGraphBuilder(weighted=True, include_self_edges=False, normalize_self_edges=False)
        clusterer = NetworkXLabelGraphClusterer(graph_builder, method='louvain')
        classifier = LabelSpacePartitioningClassifier(
            classifier = ClassifierChain(classifier=GaussianNB()),
            clusterer = clusterer
        )
        classifier.fit(X_train, y_train)
        prediction = classifier.predict(X_test)


    For more use cases see `the label relations exploration guide <../labelrelations.ipynb>`_.

    """

    def __init__(
        self, weighted=None, include_self_edges=None, normalize_self_edges=None
    ):
        super(LabelCooccurrenceGraphBuilder, self).__init__()

        if weighted not in [True, False]:
            raise ValueError("Weighted needs to be a boolean")

        if include_self_edges not in [True, False]:
            raise ValueError(
                "Decision whether to include self edges needs to be a boolean"
            )

        if include_self_edges and (normalize_self_edges not in [True, False]):
            raise ValueError(
                "Decision whether to normalize self edges needs to be a boolean"
            )

        if normalize_self_edges and not include_self_edges:
            raise ValueError(
                "Include self edges must be set to true if normalization is true"
            )

        if normalize_self_edges and not weighted:
            raise ValueError(
                "Normalizing self-edge weights_ does not make sense in an unweighted graph"
            )

        self.is_weighted = weighted
        self.include_self_edges = include_self_edges
        self.normalize_self_edges = normalize_self_edges

    def transform(self, y):
        """Generate adjacency matrix from label matrix

        This function generates a weighted or unweighted co-occurence Label Graph adjacency matrix in dictionary of keys
        format based on input binary label vectors

        Parameters
        ----------
        y : numpy.ndarray or scipy.sparse
            dense or sparse binary matrix with shape
            :code:`(n_samples, n_labels)`

        Returns
        -------
        Dict[(int, int), float]
            weight map with a tuple of label indexes as keys and a the number of samples in which the two co-occurred
        """
        label_data = get_matrix_in_format(y, "lil")
        label_count = label_data.shape[1]
        edge_map = {}

        for row in label_data.rows:
            if self.include_self_edges:
                pairs = [(a, b) for b in row for a in row if a <= b]
            else:
                pairs = [(a, b) for b in row for a in row if a < b]

            for p in pairs:
                if p not in edge_map:
                    edge_map[p] = 1.0
                else:
                    if self.is_weighted:
                        edge_map[p] += 1.0

        if self.normalize_self_edges:
            for i in range(label_count):
                if (i, i) in edge_map:
                    edge_map[(i, i)] = edge_map[(i, i)] / 2.0

        return edge_map



from __future__ import absolute_import

import numpy as np

from .base import LabelSpaceClustererBase
from .helpers import _membership_to_list_of_communities


class MatrixLabelSpaceClusterer(LabelSpaceClustererBase):
    """Cluster the label space using a scikit-compatible matrix-based clusterer

    Parameters
    ----------
    clusterer : sklearn.base.ClusterMixin
        a clonable instance of a scikit-compatible clusterer, will be automatically
        put under :code:`self.clusterer`.
    pass_input_space : bool (default is False)
        whether to take :code:`X` into consideration upon clustering,
        use only if you know that the clusterer can handle two
        parameters for clustering, will be automatically
        put under :code:`self.pass_input_space`.


    Example code for using this clusterer looks like this:

    .. code-block:: python

        from sklearn.ensemble import RandomForestClassifier
        from sklearn.cluster import KMeans
        from skmultilearn.problem_transform import LabelPowerset
        from skmultilearn.cluster import MatrixLabelSpaceClusterer
        from skmultilearn.ensemble import LabelSpacePartitioningClassifier

        # construct base forest classifier
        base_classifier = RandomForestClassifier(n_estimators=1030)

        # setup problem transformation approach with sparse matrices for random forest
        problem_transform_classifier = LabelPowerset(classifier=base_classifier,
            require_dense=[False, False])

        # setup the clusterer
        clusterer = MatrixLabelSpaceClusterer(clusterer=KMeans(n_clusters=3))

        # setup the ensemble metaclassifier
        classifier = LabelSpacePartitioningClassifier(problem_transform_classifier, clusterer)

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)
    """

    def __init__(self, clusterer=None, pass_input_space=False):
        super(MatrixLabelSpaceClusterer, self).__init__()

        self.clusterer = clusterer
        self.pass_input_space = pass_input_space

    def fit_predict(self, X, y):
        """Clusters the output space

        The clusterer's :code:`fit_predict` method is executed
        on either X and y.T vectors (if :code:`self.pass_input_space` is true)
        or just y.T to detect clusters of labels.

        The transposition of label space is used to align with
        the format expected by scikit-learn classifiers, i.e. we cluster
        labels with label assignment vectors as samples.

        Returns
        -------
        arrray of arrays of label indexes (numpy.ndarray)
            label space division, each sublist represents labels that are in that community
        """

        if self.pass_input_space:
            result = self.clusterer.fit_predict(X, y.transpose())
        else:
            result = self.clusterer.fit_predict(y.transpose())
        return np.array(_membership_to_list_of_communities(result, 1 + max(result)), dtype=object)



import numpy as np
from ..utils import get_matrix_in_format, matrix_creation_function_for_format
from scipy.sparse import issparse
from sklearn.base import BaseEstimator, ClassifierMixin


class MLClassifierBase(BaseEstimator, ClassifierMixin):
    """Base class providing API and common functions for all multi-label
    classifiers.

    Implements base functionality for ML classifiers, especially the get/set params for
    scikit-learn compatibility.

    Attributes
    ----------
    copyable_attrs : List[str]
        list of attribute names that should be copied when class is cloned
    """

    def __init__(self):
        super(MLClassifierBase, self).__init__()

        self.copyable_attrs = []

    def _generate_data_subset(self, y, subset, axis):
        """Subset rows or columns from matrix

        This function subsets the array of binary label vectors to
        include only certain labels.

        Parameters
        ----------
        y : array-like of array-likes
            An array-like of binary label vectors.
        subset: array-like of integers
            array of integers, indices that will be subsetted from
            array-likes in y
        axis: integer 0 for 'rows', 1 for 'labels',
            control variable for whether to return rows or labels as
            indexed by subset

        Returns
        -------
        multi-label binary label vector : array-like of array-likes of {0,1}
            array of binary label vectors including label data only for
            labels from parameter labels
        """
        return_data = None
        if axis == 1:
            return_data = y.tocsc()[:, subset]
        elif axis == 0:
            return_data = y.tocsr()[subset, :]

        return return_data

    def _ensure_input_format(self, X, sparse_format="csr", enforce_sparse=False):
        """Ensure the desired input format

        This function ensures that input format follows the
        density/sparsity requirements of base classifier.

        Parameters
        ----------
        X : array-like or sparse matrix
            An input feature matrix of shape :code:`(n_samples, n_features)`
        sparse_format: str
            Requested format of returned scipy.sparse matrix, if sparse is returned
        enforce_sparse : bool
            Ignore require_dense and enforce sparsity, useful internally

        Returns
        -------
        array-like or sparse matrix
            Transformed X values of shape :code:`(n_samples, n_features)`

        .. note:: If :code:`require_dense` was set to :code:`True` for
            input features in the constructor, the returned value is an
            array-like of array-likes. If :code:`require_dense` is
            set to :code:`false`, a sparse matrix of format
            :code:`sparse_format` is returned, if possible - without cloning.
        """
        is_sparse = issparse(X)

        if is_sparse:
            if self.require_dense[0] and not enforce_sparse:
                return X.toarray()
            else:
                if sparse_format is None:
                    return X
                else:
                    return get_matrix_in_format(X, sparse_format)
        else:
            if self.require_dense[0] and not enforce_sparse:
                # TODO: perhaps a check_array?
                return X
            else:
                return matrix_creation_function_for_format(sparse_format)(X)

    def _ensure_output_format(self, matrix, sparse_format="csr", enforce_sparse=False):
        """Ensure the desired output format

        This function ensures that output format follows the
        density/sparsity requirements of base classifier.

        Parameters
        ----------

        matrix : array-like matrix
            An input feature matrix of shape :code:`(n_samples)` or
            :code:`(n_samples, n_outputs)` or a sparse matrix of shape
            :code:`(n_samples, n_outputs)`

        sparse_format: str (default is csr)
            Requested format of returned :code:`scipy.sparse` matrix,
            if sparse is returned

        enforce_sparse : bool (default is False)
            Ignore :code:`require_dense` and enforce sparsity, useful
            internally

        Returns
        -------
        array-like or sparse matrix
            Transformed X values of shape :code:`(n_samples, n_features)`

        .. note:: If :code:`require_dense` was set to :code:`True` for
            input features in the constructor, the returned value is an
            array-like of array-likes. If :code:`require_dense` is
            set to :code:`false`, a sparse matrix of format
            :code:`sparse_format` is returned, if possible - without cloning.
        """
        is_sparse = issparse(matrix)

        if is_sparse:
            if self.require_dense[1] and not enforce_sparse:
                if matrix.shape[1] != 1:
                    return matrix.toarray()
                elif matrix.shape[1] == 1:
                    return np.ravel(matrix.toarray())
            else:
                if sparse_format is None:
                    return matrix
                else:
                    return get_matrix_in_format(matrix, sparse_format)
        else:
            if self.require_dense[1] and not enforce_sparse:
                # ensuring 1d
                if len(matrix.shape) > 1:
                    # a regular dense np.matrix or np.array of np.arrays
                    return np.ravel(matrix)
                else:
                    return matrix
            else:
                # ensuring 2d
                if len(matrix.shape) == 1:
                    matrix = matrix.reshape((matrix.shape[0], 1))
                return matrix_creation_function_for_format(sparse_format)(matrix)

    def fit(self, X, y):
        """Abstract method to fit classifier with training data

        It must return a fitted instance of :code:`self`.

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features, can be a dense or sparse matrix of size
            :code:`(n_samples, n_features)`
        y : numpy.ndaarray or scipy.sparse {0,1}
            binary indicator matrix with label assignments.

        Returns
        -------
        object
            fitted instance of self

        Raises
        ------
        NotImplementedError
            this is just an abstract method
        """

        raise NotImplementedError("MLClassifierBase::fit()")

    def predict(self, X):
        """Abstract method to predict labels

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of int
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`

        Raises
        ------
        NotImplementedError
            this is just an abstract method
        """
        raise NotImplementedError("MLClassifierBase::predict()")

    def get_params(self, deep=True):
        """Get parameters to sub-objects

        Introspection of classifier for search models like
        cross-validation and grid search.

        Parameters
        ----------
        deep : bool
            if :code:`True` all params will be introspected also and
            appended to the output dictionary.

        Returns
        -------
        out : dict
            dictionary of all parameters and their values. If
            :code:`deep=True` the dictionary also holds the parameters
            of the parameters.
        """
        out = dict()

        for attr in self.copyable_attrs:
            out[attr] = getattr(self, attr)

            if hasattr(getattr(self, attr), "get_params") and deep:
                deep_items = list(getattr(self, attr).get_params().items())
                out.update((attr + "__" + k, val) for k, val in deep_items)

        return out

    def set_params(self, **parameters):
        """Propagate parameters to sub-objects

        Set parameters as returned by :code:`get_params`. Please
        see this `link`_.

        .. _link: https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/base.py#L243
        """

        if not parameters:
            return self

        valid_params = self.get_params(deep=True)

        parameters_current_level = [x for x in parameters if "__" not in x]
        for parameter in parameters_current_level:
            value = parameters[parameter]

            if parameter in valid_params:
                setattr(self, parameter, value)
            else:
                raise ValueError(
                    "Invalid parameter %s for estimator %s. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`." % (parameter, self)
                )

        parameters_below_current_level = [x for x in parameters if "__" in x]
        parameters_grouped_by_current_level = {object: {} for object in valid_params}

        for parameter in parameters_below_current_level:
            object_name, sub_param = parameter.split("__", 1)

            if object_name not in parameters_grouped_by_current_level:
                raise ValueError(
                    "Invalid parameter %s for estimator %s. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`." % (object_name, self)
                )

            value = parameters[parameter]
            parameters_grouped_by_current_level[object_name][sub_param] = value

        valid_params = self.get_params(deep=True)

        # parameters_grouped_by_current_level groups valid parameters for subojects
        for object_name, sub_params in parameters_grouped_by_current_level.items():
            if len(sub_params) > 0:
                sub_object = valid_params[object_name]
                sub_object.set_params(**sub_params)

        return self



from __future__ import absolute_import

import random

import numpy as np

from .base import LabelSpaceClustererBase


class RandomLabelSpaceClusterer(LabelSpaceClustererBase):
    """Randomly divides the label space into equally-sized clusters

    This method divides the label space by drawing without replacement a desired number of
    equally sized subsets of label space, in a partitioning or overlapping scheme.

    Parameters
    ----------
    cluster_size : int
        desired size of a single cluster, will be automatically
        put under :code:`self.cluster_size`.
    cluster_count: int
        number of clusters to divide into, will be automatically
        put under :code:`self.cluster_count`.
    allow_overlap : bool
        whether to allow overlapping clusters or not, will be automatically
        put under :code:`self.allow_overlap`.

    Examples
    --------

    The following code performs random label space partitioning.

    .. code :: python

        from skmultilearn.cluster import RandomLabelSpaceClusterer

        # assume X,y contain the data, example y contains 5 labels
        cluster_count = 2
        cluster_size = y.shape[1]//cluster_count # == 2
        clr = RandomLabelSpaceClusterer(cluster_size, cluster_count, allow_overlap=False)
        clr.fit_predict(X,y)
        # Result:
        # array([list([0, 4]), list([2, 3]), list([1])], dtype=object)


    Note that the leftover labels that did not fit in `cluster_size` x `cluster_count` classifiers will be appended
    to an additional last cluster of size at most `cluster_size` - 1.

    You can also use this class to get a random division of the label space, even with multiple overlaps:

    .. code :: python

        from skmultilearn.cluster import RandomLabelSpaceClusterer

        cluster_size = 3
        cluster_count = 5
        clr = RandomLabelSpaceClusterer(cluster_size, cluster_count, allow_overlap=True)
        clr.fit_predict(X,y)

        # Result
        # array([[2, 1, 3],
        #        [3, 0, 4],
        #        [2, 3, 1],
        #        [2, 3, 4],
        #        [3, 4, 0],
        #        [3, 0, 2]])


    Note that you will never get the same label subset twice.
    """

    def __init__(self, cluster_size, cluster_count, allow_overlap):
        super(RandomLabelSpaceClusterer, self).__init__()

        self.cluster_size = cluster_size
        self.cluster_count = cluster_count
        self.allow_overlap = allow_overlap

    def fit_predict(self, X, y):
        """Cluster the output space

        Parameters
        ----------
        X : currently unused, left for scikit compatibility
        y : scipy.sparse
            label space of shape :code:`(n_samples, n_labels)`

        Returns
        -------
        arrray of arrays of label indexes (numpy.ndarray)
            label space division, each sublist represents labels that are in that community
        """

        if (self.cluster_count + 1) * self.cluster_size < y.shape[1]:
            raise ValueError(
                "Cannot include all of {} labels in {} clusters of {} labels".format(
                    y.shape[1], self.cluster_count, self.cluster_size
                )
            )

        all_labels_assigned_to_division = False
        # make sure the final label set division includes all labels
        while not all_labels_assigned_to_division:
            label_sets = []
            free_labels = range(y.shape[1])

            while len(label_sets) <= self.cluster_count:
                if not self.allow_overlap:
                    if len(free_labels) == 0:
                        break

                    # in this case, we are unable to draw new labels, add all that remain
                    if len(free_labels) < self.cluster_size:
                        label_sets.append(free_labels)
                        break

                label_set = random.sample(free_labels, self.cluster_size)
                if not self.allow_overlap:
                    free_labels = list(set(free_labels).difference(set(label_set)))

                if label_set not in label_sets:
                    label_sets.append(label_set)

            all_labels_assigned_to_division = all(
                any(label in subset for subset in label_sets)
                for label in range(y.shape[1])
            )

        return np.array(label_sets, dtype=object)



from skmultilearn.base import ProblemTransformationBase
import numpy as np
import scipy.sparse as sp
from copy import copy


class EmbeddingClassifier(ProblemTransformationBase):
    """Embedding-based classifier

    Implements a general scheme presented in LNEMLC: label network embeddings for multi-label classification. The
    classifier embeds the label space with the embedder, trains a set of single-variate or a multi-variate regressor
    for embedding unseen cases and a base classifier to predict labels based on input features and the embeddings.

    Parameters
    ----------
    embedder : :class:`~sklearn.base.BaseEstimator`
        the class to embed the label space

    regressor : :class:`~sklearn.base.BaseEstimator`
        the base regressor to predict embeddings from input features

    classifier : :class:`~sklearn.base.BaseEstimator`
        the base classifier to predict labels from input features and embeddings

    regressor_per_dimension : bool
        whether to train one joint multi-variate regressor (False) or per dimension single-variate regressor (True)

    require_dense : [bool, bool], optional
        whether the base classifier requires dense representations for input features and classes/labels
        matrices in fit/predict.


    Attributes
    ----------
    n_regressors_ : int
        number of trained regressors
    partition_ : List[List[int]], shape=(`model_count_`,)
        list of lists of label indexes, used to index the output space matrix, set in :meth:`_generate_partition`
        via :meth:`fit`
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
        list of classifiers trained per partition, set in :meth:`fit`


    If you use this classifier please cite the relevant embedding method paper
    and the label network embedding for multi-label classification paper:

    .. code :: bibtex

        @article{zhang2007ml,
          title={ML-KNN: A lazy learning approach to multi-label learning},
          author={Zhang, Min-Ling and Zhou, Zhi-Hua},
          journal={Pattern recognition},
          volume={40},
          number={7},
          pages={2038--2048},
          year={2007},
          publisher={Elsevier}
        }

    Example
    -------
    An example use case for EmbeddingClassifier:

    .. code-block:: python

        from skmultilearn.embedding import SKLearnEmbedder, EmbeddingClassifier
        from sklearn.manifold import SpectralEmbedding
        from sklearn.ensemble import RandomForestRegressor
        from skmultilearn.adapt import MLkNN

        clf = EmbeddingClassifier(
            SKLearnEmbedder(SpectralEmbedding(n_components = 10)),
            RandomForestRegressor(n_estimators=10),
            MLkNN(k=5)
        )

        clf.fit(X_train, y_train)

        predictions = clf.predict(X_test)

    """

    def __init__(
        self,
        embedder,
        regressor,
        classifier,
        regressor_per_dimension=False,
        require_dense=None,
    ):
        super(EmbeddingClassifier, self).__init__()
        self.embedder = embedder
        self.regressor = regressor
        self.classifier = classifier
        self.regressor_per_dimension = regressor_per_dimension

        if require_dense is None:
            require_dense = [True, True]

        self.require_dense = require_dense

        self.copyable_attrs = [
            "embedder",
            "regressor",
            "classifier",
            "regressor_per_dimension",
            "require_dense",
        ]

    def fit(self, X, y):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self
        """

        X = self._ensure_input_format(X)
        y = self._ensure_input_format(y)

        y_embedded = self.embedder.fit_transform(X, y)[1]
        X_y_embedded = self._concatenate_matrices(X, y_embedded)

        if self.regressor_per_dimension:
            self.n_regressors_ = y_embedded.shape[1]
            self.regressors_ = [None for _ in range(self.n_regressors_)]
            for i in range(self.n_regressors_):
                self.regressors_[i] = copy(self.regressor)
                self.regressors_[i].fit(X, y_embedded[:, i])
        else:
            self.n_regressors_ = 1

            self.regressor.fit(X, y_embedded)
        self.classifier.fit(X_y_embedded, y)

        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """

        X = self._ensure_input_format(X)

        X_y_embedded = self._predict_embedding(X)
        return self.classifier.predict(X_y_embedded)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """

        X_y_embedded = self._predict_embedding(X)
        return self.classifier.predict_proba(X_y_embedded)

    def _concatenate_matrices(self, X, y_embedded):
        X = self._ensure_input_format(X)
        y_embedded = self._ensure_input_format(y_embedded)

        if sp.issparse(X):
            X_y_embedded = sp.hstack([X, y_embedded])
        else:
            X_y_embedded = np.hstack([X, y_embedded])

        return X_y_embedded

    def _predict_embedding(self, X):
        if self.regressor_per_dimension:
            y_embedded = [
                self.regressors_[i].predict(X) for i in range(self.n_regressors_)
            ]
            if sp.issparse(X):
                y_embedded = sp.csr_matrix(y_embedded).T
            else:
                y_embedded = np.matrix(y_embedded).T
        else:
            y_embedded = self.regressor.predict(X)

        return self._concatenate_matrices(X, y_embedded)



import os
import shlex
import subprocess
import sys
import tempfile
import zipfile
from builtins import filter
from builtins import map
from builtins import range
from builtins import str

import scipy.sparse as sparse

from ..base import MLClassifierBase
from ..dataset import save_to_arff, get_data_home, _download_single_file, _get_md5

try:
    from shlex import quote as cmd_quote
except ImportError:
    from pipes import quote as cmd_quote

SUPPORTED_VERSION = "1.9.2"
SUPPORTED_VERSION_MD5 = "e909044b39513bbad451b8d71098b22c"


def download_meka(version=None):
    """Downloads a given version of the MEKA library and returns its classpath

    Parameters
    ----------
    version : str
        the MEKA version to download, default falls back to currently supported version 1.9.2

    Returns
    -------
    string
        meka class path string for installed version

    Raises
    ------
    IOError
        if unpacking the meka release file does not provide a proper setup
    Exception
        if MD5 mismatch happens after a download error
    """
    version = version or SUPPORTED_VERSION
    meka_release_string = "meka-release-{}".format(version)
    file_name = meka_release_string + "-bin.zip"
    meka_path = get_data_home(subdirectory="meka")
    target_path = os.path.join(meka_path, file_name)
    path_to_lib = os.path.join(meka_path, meka_release_string, "lib")

    if os.path.exists(target_path):
        print("MEKA {} found, not downloading".format(version))

    else:
        print("MEKA {} not found, downloading".format(version))
        release_url = "http://downloads.sourceforge.net/project/meka/meka-{}/".format(
            version
        )
        _download_single_file(file_name, target_path, release_url)

        found_md5 = _get_md5(target_path)
        if SUPPORTED_VERSION_MD5 != found_md5:
            raise Exception("MD5 mismatch - possible MEKA download error")

    if not os.path.exists(path_to_lib):
        with zipfile.ZipFile(target_path, "r") as meka_zip:
            print("Unzipping MEKA {} to {}".format(version, meka_path + os.path.sep))
            meka_zip.extractall(path=meka_path + os.path.sep)

    if not os.path.exists(os.path.join(path_to_lib, "meka-{}.jar".format(version))):
        raise IOError(
            "Something went wrong, MEKA files missing, please file a bug report"
        )

    return path_to_lib + os.path.sep


class Meka(MLClassifierBase):
    """Wrapper for the MEKA classifier

    Allows using MEKA, WEKA and some of MULAN classifiers from scikit-compatible API. For more information on
    how to use this class see the tutorial: :doc:`../meka`

    Parameters
    ----------
    meka_classifier : str
        The MEKA classifier string and parameters from the MEKA API,
        such as :code:`meka.classifiers.multilabel.MULAN -S RAkEL2`
    weka_classifier : str
        The WEKA classifier string and parameters from the WEKA API,
        such as :code:`weka.classifiers.trees.J48`
    java_command : str
        Path to test the java command
    meka_classpath: str
        Path to the MEKA class path folder, usually the folder lib
        in the directory MEKA was extracted into

    Attributes
    ----------
    output_ : str
        the full text output of MEKA command

    References
    ----------

    If you use this wrapper please also cite:

    .. code-block :: latex

        @article{MEKA,
            author = {Read, Jesse and Reutemann, Peter and Pfahringer, Bernhard and Holmes, Geoff},
            title = {{MEKA}: A Multi-label/Multi-target Extension to {Weka}},
            journal = {Journal of Machine Learning Research},
            year = {2016},
            volume = {17},
            number = {21},
            pages = {1--5},
            url = {http://jmlr.org/papers/v17/12-164.html},
        }

        @article{Hall:2009:WDM:1656274.1656278,
            author = {Hall, Mark and Frank, Eibe and Holmes, Geoffrey and Pfahringer, Bernhard and Reutemann, Peter and Witten, Ian H.},
            title = {The WEKA Data Mining Software: An Update},
            journal = {SIGKDD Explor. Newsl.},
            issue_date = {June 2009},
            volume = {11},
            number = {1},
            month = nov,
            year = {2009},
            issn = {1931-0145},
            pages = {10--18},
            numpages = {9},
            url = {http://doi.acm.org/10.1145/1656274.1656278},
            doi = {10.1145/1656274.1656278},
            acmid = {1656278},
            publisher = {ACM},
            address = {New York, NY, USA},
        }

    Examples
    --------

    Here's an example of performing Label Powerset classification using MEKA with a WEKA Naive Bayes classifier.

    .. code-block:: python

        from skmultilearn.ext import Meka, download_meka

        meka = Meka(
            meka_classifier = "meka.classifiers.multilabel.LC",
            weka_classifier = "weka.classifiers.bayes.NaiveBayes",
            meka_classpath = download_meka(),
            java_command = '/usr/bin/java')

        meka.fit(X_train, y_train)
        predictions = meka.predict(X_test)

    """

    def __init__(
        self,
        meka_classifier=None,
        weka_classifier=None,
        java_command=None,
        meka_classpath=None,
    ):
        super(Meka, self).__init__()

        self.java_command = java_command
        if self.java_command is None:
            # TODO: this will not be needed once we're python 3 ready - we will
            # use it only in python 2.7 cases
            from whichcraft import which

            self.java_command = which("java")

            if self.java_command is None:
                raise ValueError("Java not found")

        self.meka_classpath = meka_classpath
        if self.meka_classpath is None:
            self.meka_classpath = os.environ.get("MEKA_CLASSPATH")

            if self.meka_classpath is None:
                raise ValueError("No meka classpath defined")

        self.meka_classifier = meka_classifier
        self.weka_classifier = weka_classifier

        self.copyable_attrs = [
            "meka_classifier",
            "weka_classifier",
            "java_command",
            "meka_classpath",
        ]
        self.output_ = None
        self._verbosity = 5
        self._warnings = None
        self.require_dense = [False, False]

        self._clean()

    def _clean(self):
        """Sets various attributes to :code:`None`"""
        self._results = None
        self._statistics = None
        self.output_ = None
        self._error = None
        self._label_count = None
        self._instance_count = None

    def _remove_temporary_files(self, temporary_files):
        """Internal function for cleaning temporary files"""
        for file_object in temporary_files:
            file_name = file_object.name
            file_object.close()
            if os.path.exists(file_name):
                os.remove(file_name)

            arff_file_name = file_name + ".arff"
            if os.path.exists(arff_file_name):
                os.remove(arff_file_name)

    def fit(self, X, y):
        """Fits classifier to training data

        Internally this method dumps X and y to temporary arff files and
        runs MEKA with relevant arguments using :meth:`_run`. It uses a
        sparse DOK representation (:class:`scipy.sparse.dok_matrix`)
        of the X matrix.

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self
        """
        self._clean()
        X = self._ensure_input_format(X, sparse_format="dok", enforce_sparse=True)
        y = self._ensure_output_format(y, sparse_format="dok", enforce_sparse=True)
        self._label_count = y.shape[1]

        # we need this in case threshold needs to be recalibrated in meka
        self.train_data_ = save_to_arff(X, y)
        train_arff = tempfile.NamedTemporaryFile(delete=False)
        classifier_dump_file = tempfile.NamedTemporaryFile(delete=False)

        try:
            with open(train_arff.name + ".arff", "w") as fp:
                fp.write(self.train_data_)

            input_args = [
                "-verbosity",
                "0",
                "-split-percentage",
                "100",
                "-t",
                '"{}"'.format(train_arff.name + ".arff"),
                "-d",
                '"{}"'.format(classifier_dump_file.name),
            ]

            self._run_meka_command(input_args)
            self.classifier_dump = None
            with open(classifier_dump_file.name, "rb") as fp:
                self.classifier_dump = fp.read()
        finally:
            self._remove_temporary_files([train_arff, classifier_dump_file])

        return self

    def predict(self, X):
        """Predict label assignments for X

        Internally this method dumps X to temporary arff files and
        runs MEKA with relevant arguments using :func:`_run`. It uses a
        sparse DOK representation (:class:`scipy.sparse.dok_matrix`)
        of the X matrix.

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of int
            sparse matrix of integers with shape :code:`(n_samples, n_features)`
        """
        X = self._ensure_input_format(X, sparse_format="dok", enforce_sparse=True)
        self._instance_count = X.shape[0]

        if self.classifier_dump is None:
            raise Exception("Not classified")

        sparse_y = sparse.coo_matrix((X.shape[0], self._label_count), dtype=int)

        try:
            train_arff = tempfile.NamedTemporaryFile(delete=False)
            test_arff = tempfile.NamedTemporaryFile(delete=False)
            classifier_dump_file = tempfile.NamedTemporaryFile(delete=False)

            with open(train_arff.name + ".arff", "w") as fp:
                fp.write(self.train_data_)

            with open(classifier_dump_file.name, "wb") as fp:
                fp.write(self.classifier_dump)

            with open(test_arff.name + ".arff", "w") as fp:
                fp.write(save_to_arff(X, sparse_y))

            args = ["-l", '"{}"'.format(classifier_dump_file.name)]

            self._run(train_arff.name + ".arff", test_arff.name + ".arff", args)
            self._parse_output()

        finally:
            self._remove_temporary_files([train_arff, test_arff, classifier_dump_file])

        return self._results

    def _run(self, train_file, test_file, additional_arguments=[]):
        """Runs the meka classifiers

        Parameters
        ----------
        train_file : str
            path to train :code:`.arff` file in meka format
            (big endian, labels first in attributes list).
        test_file : str
            path to test :code:`.arff` file in meka format
            (big endian, labels first in attributes list).

        Returns
        -------
        predictions: sparse binary indicator matrix [n_test_samples, n_labels]
            array of binary label vectors including label predictions of
            shape :code:`(n_test_samples, n_labels)`
        """
        self.output_ = None
        self._warnings = None

        # meka_command_string = 'java -cp "/home/niedakh/pwr/old/meka-1.5/lib/*" meka.classifiers.multilabel.MULAN -S RAkEL2
        # -threshold 0 -t {train} -T {test} -verbosity {verbosity} -W weka.classifiers.bayes.NaiveBayes'
        # meka.classifiers.multilabel.LC, weka.classifiers.bayes.NaiveBayes

        args = [
            "-t",
            '"{}"'.format(train_file),
            "-T",
            '"{}"'.format(test_file),
            "-verbosity",
            str(5),
        ] + additional_arguments

        self._run_meka_command(args)
        return self

    def _parse_output(self):
        """Internal function for parsing MEKA output."""
        if self.output_ is None:
            self._results = None
            self._statistics = None
            return None

        predictions_split_head = "==== PREDICTIONS"
        predictions_split_foot = "|==========="

        if self._label_count is None:
            self._label_count = map(
                lambda y: int(y.split(")")[1].strip()),
                [x for x in self.output_.split("\n") if "Number of labels" in x],
            )[0]

        if self._instance_count is None:
            self._instance_count = int(
                float(
                    filter(
                        lambda x: "==== PREDICTIONS (N=" in x, self.output_.split("\n")
                    )[0]
                    .split("(")[1]
                    .split("=")[1]
                    .split(")")[0]
                )
            )
        predictions = (
            self.output_.split(predictions_split_head)[1]
            .split(predictions_split_foot)[0]
            .split("\n")[1:-1]
        )

        predictions = [
            y.split("]")[0] for y in [x.split("] [")[1] for x in predictions]
        ]
        predictions = [
            [a for a in [f.strip() for f in z.split(",")] if len(a) > 0]
            for z in predictions
        ]
        predictions = [[int(a) for a in z] for z in predictions]

        assert self._verbosity == 5

        self._results = sparse.lil_matrix(
            (self._instance_count, self._label_count), dtype="int"
        )
        for row in range(self._instance_count):
            for label in predictions[row]:
                self._results[row, label] = 1

        statistics = [
            x
            for x in self.output_.split("== Evaluation Info")[1].split("\n")
            if len(x) > 0 and "==" not in x
        ]
        statistics = [y for y in [z.strip() for z in statistics] if "  " in y]
        array_data = [z for z in statistics if "[" in z]
        non_array_data = [z for z in statistics if "[" not in z]

        self._statistics = {}
        for row in non_array_data:
            r = row.strip().split("  ")
            r = [z for z in r if len(z) > 0]
            r = [z.strip() for z in r]
            if len(r) < 2:
                continue
            try:
                test_value = float(r[1])
            except ValueError:
                test_value = r[1]

            r[1] = test_value
            self._statistics[r[0]] = r[1]

        for row in array_data:
            r = row.strip().split("[")
            r = [z.strip() for z in r]
            r[1] = r[1].replace(", ", " ").replace(",", ".").replace("]", "").split(" ")
            r[1] = [x for x in r[1] if len(x) > 0]
            self._statistics[r[0]] = r[1]

    def _run_meka_command(self, args):
        command_args = [
            self.java_command,
            "-cp",
            '"{}*"'.format(self.meka_classpath),
            self.meka_classifier,
        ]
        
        command_args += args

        if self.weka_classifier is not None:
            command_args += ["-W", self.weka_classifier]

        meka_command = " ".join(command_args)

        if sys.platform != "win32":
            meka_command = shlex.split(meka_command)

        pipes = subprocess.Popen(
            meka_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        self.output_, self._error = pipes.communicate()
        if type(self.output_) == bytes:
            self.output_ = self.output_.decode(sys.stdout.encoding)
        if type(self._error) == bytes:
            self._error = self._error.decode(sys.stdout.encoding)

        if pipes.returncode != 0:
            raise Exception(self.output_ + self._error)



from ..base.problem_transformation import ProblemTransformationBase
from scipy.sparse import hstack
from sklearn.exceptions import NotFittedError
import copy


class ClassifierChain(ProblemTransformationBase):
    """Constructs a bayesian conditioned chain of per label classifiers

    This class provides implementation of Jesse Read's problem
    transformation method called Classifier Chains. For L labels it
    trains L classifiers ordered in a chain according to the
    `Bayesian chain rule`.

    The first classifier is trained just on the input space, and then
    each next classifier is trained on the input space and all previous
    classifiers in the chain.

    The default classifier chains follow the same ordering as provided
    in the training set, i.e. label in column 0, then 1, etc.

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        scikit-learn compatible base classifier
    require_dense : [bool, bool], optional
        whether the base classifier requires dense representations
        for input features and classes/labels matrices in fit/predict.
        If value not provided, sparse representations are used if base classifier is
        an instance of :class:`~skmultilearn.base.MLClassifierBase` and dense otherwise.
    order : List[int], permutation of ``range(n_labels)``, optional
        the order in which the chain should go through labels, the default is ``range(n_labels)``


    Attributes
    ----------
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `n_labels`
        list of classifiers trained per partition, set in :meth:`fit`



    References
    ----------
    If used, please cite the scikit-multilearn library and the relevant paper:

    .. code-block:: bibtex

        @inproceedings{read2009classifier,
          title={Classifier chains for multi-label classification},
          author={Read, Jesse and Pfahringer, Bernhard and Holmes, Geoff and Frank, Eibe},
          booktitle={Joint European Conference on Machine Learning and Knowledge Discovery in Databases},
          pages={254--269},
          year={2009},
          organization={Springer}
        }

    Examples
    --------
    An example use case for Classifier Chains
    with an :class:`sklearn.svm.SVC` base classifier which supports sparse input:

    .. code-block:: python

        from skmultilearn.problem_transform import ClassifierChain
        from sklearn.svm import SVC

        # initialize Classifier Chain multi-label classifier
        # with an SVM classifier
        # SVM in scikit only supports the X matrix in sparse representation

        classifier = ClassifierChain(
            classifier = SVC(),
            require_dense = [False, True]
        )

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    Another way to use this classifier is to select the best scenario from a set of single-label classifiers used
    with Classifier Chain, this can be done using cross validation grid search. In the example below, the model
    with highest accuracy results is selected from either a :class:`sklearn.naive_bayes.MultinomialNB` or
    :class:`sklearn.svm.SVC` base classifier, alongside with best parameters for that base classifier.

    .. code-block:: python

        from skmultilearn.problem_transform import ClassifierChain
        from sklearn.model_selection import GridSearchCV
        from sklearn.naive_bayes import MultinomialNB
        from sklearn.svm import SVC

        parameters = [
            {
                'classifier': [MultinomialNB()],
                'classifier__alpha': [0.7, 1.0],
            },
            {
                'classifier': [SVC()],
                'classifier__kernel': ['rbf', 'linear'],
            },
        ]


        clf = GridSearchCV(ClassifierChain(), parameters, scoring='accuracy')
        clf.fit(x, y)

        print (clf.best_params_, clf.best_score_)

        # result
        # {'classifier': MultinomialNB(alpha=0.7, class_prior=None, fit_prior=True), 'classifier__alpha': 0.7} 0.16

    """

    def __init__(self, classifier=None, require_dense=None, order=None):
        super(ClassifierChain, self).__init__(classifier, require_dense)
        self.order = order
        self.copyable_attrs = ["classifier", "require_dense", "order"]

    def fit(self, X, y, order=None):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """

        # fit L = len(y[0]) BR classifiers h_i
        # on X + y[:i] as input space and y[i+1] as output

        X_extended = self._ensure_input_format(
            X, sparse_format="csc", enforce_sparse=True
        )
        y = self._ensure_output_format(y, sparse_format="csc", enforce_sparse=True)

        self._label_count = y.shape[1]
        self.classifiers_ = [None for x in range(self._label_count)]

        for label in self._order():
            self.classifier = copy.deepcopy(self.classifier)
            y_subset = self._generate_data_subset(y, label, axis=1)

            self.classifiers_[label] = self.classifier.fit(
                self._ensure_input_format(X_extended),
                self._ensure_output_format(y_subset),
            )
            X_extended = hstack([X_extended, y_subset])

        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """

        X_extended = self._ensure_input_format(
            X, sparse_format="csc", enforce_sparse=True
        )

        for label in self._order():
            prediction = self.classifiers_[label].predict(
                self._ensure_input_format(X_extended)
            )
            prediction = self._ensure_multi_label_from_single_class(prediction)
            X_extended = hstack([X_extended, prediction])
        return X_extended[:, -self._label_count :]

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """
        X_extended = self._ensure_input_format(
            X, sparse_format="csc", enforce_sparse=True
        )

        results = []
        for label in self._order():
            prediction = self.classifiers_[label].predict(
                self._ensure_input_format(X_extended)
            )

            prediction = self._ensure_output_format(
                prediction, sparse_format="csc", enforce_sparse=True
            )

            prediction_proba = self.classifiers_[label].predict_proba(
                self._ensure_input_format(X_extended)
            )

            prediction_proba = self._ensure_output_format(
                prediction_proba, sparse_format="csc", enforce_sparse=True
            )[:, 1]

            X_extended = hstack([X_extended, prediction]).tocsc()
            results.append(prediction_proba)

        return hstack(results)

    def _order(self):
        if self.order is not None:
            return self.order

        try:
            return list(range(self._label_count))
        except AttributeError:
            raise NotFittedError("This Classifier Chain has not been fit yet")



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

# import sphinx_pypi_upload
import sys

if sys.version_info[0] < 3:
    import codecs

    with codecs.open("README.md", "r", encoding="utf-8") as f:
        readme = f.read()

else:
    import io

    with io.open("README.md", "r", encoding="utf-8") as f:
        readme = f.read()

setup(
    name="scikit-multilearn-ng",
    version="v0.0.1",
    packages=find_packages(exclude=["docs", "tests", "*.tests"]),
    install_requires=[
        "scipy>=1.1.0",
        "numpy>=1.15.1",
        "liac-arff>=2.2.1",
        "networkx>=2.1",
        "python-louvain>=0.11",
        "future>=0.16.0",
        "scikit_learn>=0.19.2",
        "requests>=2.18.4",
        "joblib",
    ],
    extras_require={
        "gpl": [
            "igraph",
        ],
        "dev": [
            "pytest>=3.3.1",
            "PyHamcrest>=1.9.0",
            "ipython",
            "ipython-genutils",
        ],
        "keras": [
            "keras<=2.12.0",
            "tensorflow",
        ],
        "meka": [
            "whichcraft>=0.4.1",
        ],
        "test": [
            "pytest",
            "pyhamcrest",
        ],
    },
    author="Piotr Szymański",
    author_email="niedakh@gmail.com",
    license="BSD",
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/scikit-multilearn-ng/scikit-multilearn-ng/",
    description="Scikit-multilearn-ng is the follow up to scikit-multilearn, a BSD-licensed library for multi-label classification that is built on top of the well-known scikit-learn ecosystem.",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



"""
The :mod:`skmultilearn.ensemble` module implements ensemble classification schemes
that construct an ensemble of base multi-label classifiers.

Currently the following ensemble classification schemes are available in scikit-multilearn:

+--------------------------------------------------------------------+----------------------------------------------------------------+
| Classifier name                                                    | Description                                                    |
+====================================================================+================================================================+
| :class:`~skmultilearn.ensemble.RakelD`                             | Distinct RAndom k-labELsets multi-label classifier             |
+--------------------------------------------------------------------+----------------------------------------------------------------+
| :class:`~skmultilearn.ensemble.RakelO`                             | Overlapping RAndom k-labELsets multi-label classifier.         |
+--------------------------------------------------------------------+----------------------------------------------------------------+
| :class:`~skmultilearn.ensemble.LabelSpacePartitioningClassifier`   | a label space partitioning classifier that trains a            |
|                                                                    | classifier per label subspace as clustered using methods       |
|                                                                    | from :mod:`skmultilearn.cluster`.                              |
+--------------------------------------------------------------------+----------------------------------------------------------------+
| :class:`~skmultilearn.ensemble.MajorityVotingClassifier`           | a label space division classifier that trains a classifier     |
|                                                                    | per label subspace as clustered using methods from             |
|                                                                    | :mod:`skmultilearn.cluster` and assign labels if the majority  |
|                                                                    | of classifiers that contain the label agree on the assignment. |
+--------------------------------------------------------------------+----------------------------------------------------------------+
"""

from __future__ import absolute_import
from .rakeld import RakelD
from .rakelo import RakelO
from .partition import LabelSpacePartitioningClassifier
from .voting import MajorityVotingClassifier

__all__ = [
    "RakelD",
    "RakelO",
    "LabelSpacePartitioningClassifier",
    "LabelSpacePartitioningClassifier",
    "MajorityVotingClassifier",
]



from ..base.problem_transformation import ProblemTransformationBase
import numpy as np
from scipy import sparse


class LabelPowerset(ProblemTransformationBase):
    """Transform multi-label problem to a multi-class problem

    Label Powerset is a problem transformation approach to multi-label
    classification that transforms a multi-label problem to a multi-class
    problem with 1 multi-class classifier trained on all unique label
    combinations found in the training data.

    The method maps each combination to a unique combination id number, and performs multi-class classification
    using the `classifier` as multi-class classifier and combination ids as classes.

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        scikit-learn compatible base classifier
    require_dense : [bool, bool], optional
        whether the base classifier requires dense representations
        for input features and classes/labels matrices in fit/predict.
        If value not provided, sparse representations are used if base classifier is
        an instance of :class:`skmultilearn.base.MLClassifierBase` and dense otherwise.

    Attributes
    ----------
    unique_combinations_ : Dict[str, int]
        mapping from label combination as string to label combination id :meth:`transform:` via :meth:`fit`
    reverse_combinations_ : List[List[int]]
        label combination id ordered list to list of label indexes for a given combination  :meth:`transform:`
        via :meth:`fit`

    Notes
    -----
    .. note ::

        `n_classes` in this document denotes the number of unique label combinations present in the training `y`
        passed to :meth:`fit`, in practice it is equal to :code:`len(self.unique_combinations)`

    Examples
    --------
    An example use case for Label Powerset with an :class:`sklearn.ensemble.RandomForestClassifier` base classifier
    which supports sparse input:

    .. code-block:: python

        from skmultilearn.problem_transform import LabelPowerset
        from sklearn.ensemble import RandomForestClassifier

        # initialize LabelPowerset multi-label classifier with a RandomForest
        classifier = LabelPowerset(
            classifier = RandomForestClassifier(n_estimators=100),
            require_dense = [False, True]
        )

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    Another way to use this classifier is to select the best scenario from a set of multi-class classifiers used
    with Label Powerset, this can be done using cross validation grid search. In the example below, the model
    with highest accuracy results is selected from either a :class:`sklearn.ensemble.RandomForestClassifier` or
    :class:`sklearn.naive_bayes.MultinomialNB` base classifier, alongside with best parameters for
    that base classifier.

    .. code-block:: python

        from skmultilearn.problem_transform import LabelPowerset
        from sklearn.model_selection import GridSearchCV
        from sklearn.naive_bayes import MultinomialNB
        from sklearn.ensemble import RandomForestClassifier

        parameters = [
            {
                'classifier': [MultinomialNB()],
                'classifier__alpha': [0.7, 1.0],
            },
            {
                'classifier': [RandomForestClassifier()],
                'classifier__criterion': ['gini', 'entropy'],
                'classifier__n_estimators': [10, 20, 50],
            },
        ]

        clf = GridSearchCV(LabelPowerset(), parameters, scoring='accuracy')
        clf.fit(x, y)

        print (clf.best_params_, clf.best_score_)

        # result
        # {
        #   'classifier': RandomForestClassifier(bootstrap=True, class_weight=None, criterion='gini',
        #             max_depth=None, max_features='auto', max_leaf_nodes=None,
        #             min_impurity_decrease=0.0, min_impurity_split=None,
        #             min_samples_leaf=1, min_samples_split=2,
        #             min_weight_fraction_leaf=0.0, n_estimators=50, n_jobs=1,
        #             oob_score=False, random_state=None, verbose=0,
        #             warm_start=False), 'classifier__criterion': 'gini', 'classifier__n_estimators': 50
        # } 0.16

    """

    def __init__(self, classifier=None, require_dense=None):
        super(LabelPowerset, self).__init__(
            classifier=classifier, require_dense=require_dense
        )
        self._clean()

    def _clean(self):
        """Reset classifier internals before refitting"""
        self.unique_combinations_ = {}
        self.reverse_combinations_ = []
        self._label_count = None

    def fit(self, X, y):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)

        self.classifier.fit(self._ensure_input_format(X), self.transform(y))

        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """

        # this will be an np.array of integers representing classes
        lp_prediction = self.classifier.predict(self._ensure_input_format(X))

        return self.inverse_transform(lp_prediction)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """

        lp_prediction = self.classifier.predict_proba(self._ensure_input_format(X))
        result = sparse.lil_matrix((X.shape[0], self._label_count), dtype="float")
        for row in range(len(lp_prediction)):
            assignment = lp_prediction[row]
            for combination_id in range(len(assignment)):
                for label in self.reverse_combinations_[combination_id]:
                    result[row, label] += assignment[combination_id]

        return result

    def transform(self, y):
        """Transform multi-label output space to multi-class

        Transforms a mutli-label problem into a single-label multi-class
        problem where each label combination is a separate class.

        Parameters
        -----------
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        numpy.ndarray of `{0, ... , n_classes-1}`, shape=(n_samples,)
            a multi-class output space vector

        """

        y = self._ensure_output_format(y, sparse_format="lil", enforce_sparse=True)

        self._clean()
        self._label_count = y.shape[1]

        last_id = 0
        train_vector = []
        for labels_applied in y.rows:
            label_string = ",".join(map(str, labels_applied))

            if label_string not in self.unique_combinations_:
                self.unique_combinations_[label_string] = last_id
                self.reverse_combinations_.append(labels_applied)
                last_id += 1

            train_vector.append(self.unique_combinations_[label_string])

        return np.array(train_vector)

    def inverse_transform(self, y):
        """Transforms multi-class assignment to multi-label

        Transforms a mutli-label problem into a single-label multi-class
        problem where each label combination is a separate class.

        Parameters
        -----------
        y : numpy.ndarray of `{0, ... , n_classes-1}`, shape=(n_samples,)
            binary indicator matrix with label assignments

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        n_samples = len(y)
        result = sparse.lil_matrix((n_samples, self._label_count), dtype="i8")
        for row in range(n_samples):
            assignment = y[row]
            result[row, self.reverse_combinations_[assignment]] = 1

        return result



"""
The :mod:`skmultilearn.problem_transform` module provides classifiers
that follow the problem transformation approaches to multi-label classification.

The problem transformation approach to multi-label classification converts multi-label problems to
single-label problems: single-class or multi-class.


+----------------------------------------------------------+------------------------------------------------+
| Classifier                                               | Description                                    |
+==========================================================+================================================+
| :class:`~skmultilearn.problem_transform.BinaryRelevance` |  treats each label as a separate single-class  |
|                                                          |  classification problem                        |
+----------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.ClassifierChain` |  treats each label as a part of a conditioned  |
|                                                          |  chain of single-class classification problems |
+----------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.LabelPowerset`   | treats each label combination as a separate    |
|                                                          | class with one multi-class classification      |
|                                                          | problem                                        |
+----------------------------------------------------------+------------------------------------------------+


"""

from .br import BinaryRelevance
from .cc import ClassifierChain
from .lp import LabelPowerset

__all__ = ["BinaryRelevance", "ClassifierChain", "LabelPowerset"]



import numpy as np

from .partition import LabelSpacePartitioningClassifier
from ..cluster.random import RandomLabelSpaceClusterer
from ..problem_transform import LabelPowerset
from ..base import MLClassifierBase


class RakelD(MLClassifierBase):
    """Distinct RAndom k-labELsets multi-label classifier.

    Divides the label space in to equal partitions of size k, trains a Label Powerset
    classifier per partition and predicts by summing the result of all trained classifiers.

    Parameters
    ----------
    base_classifier : sklearn.base
        the base classifier that will be used in a class, will be
        automatically put under :code:`self.classifier` for future
        access.
    base_classifier_require_dense : [bool, bool]
        whether the base classifier requires [input, output] matrices
        in dense representation, will be automatically
        put under :code:`self.require_dense`
    labelset_size : int
        the desired size of each of the partitions, parameter k according to paper
        Default is 3, according to paper it has the best results


    Attributes
    ----------
    _label_count : int
        the number of labels the classifier is fit to, set by :meth:`fit`

    model_count_ : int
        the number of sub classifiers trained, set by :meth:`fit`

    classifier_: :class:`skmultilearn.ensemble.LabelSpacePartitioningClassifier`
        the underneath classifier that perform the label space partitioning using a
        random clusterer :class:`skmultilearn.ensemble.RandomLabelSpaceClusterer`


    References
    ----------

    If you use this class please cite the paper introducing the method:

    .. code :: latex

        @ARTICLE{5567103,
            author={G. Tsoumakas and I. Katakis and I. Vlahavas},
            journal={IEEE Transactions on Knowledge and Data Engineering},
            title={Random k-Labelsets for Multilabel Classification},
            year={2011},
            volume={23},
            number={7},
            pages={1079-1089},
            doi={10.1109/TKDE.2010.164},
            ISSN={1041-4347},
            month={July},
        }

    Examples
    --------

    Here's a simple example of how to use this class with a base classifier from scikit-learn to teach
    non-overlapping classifiers each trained on at most four labels:

    .. code :: python

        from sklearn.naive_bayes import GaussianNB
        from skmultilearn.ensemble import RakelD

        classifier = RakelD(
            base_classifier=GaussianNB(),
            base_classifier_require_dense=[True, True],
            labelset_size=4
        )

        classifier.fit(X_train, y_train)
        prediction = classifier.predict(X_test)

    """

    def __init__(
        self, base_classifier=None, labelset_size=3, base_classifier_require_dense=None
    ):
        super(RakelD, self).__init__()

        self.labelset_size = labelset_size
        self.base_classifier = base_classifier
        self.base_classifier_require_dense = base_classifier_require_dense
        self.copyable_attrs = [
            "base_classifier",
            "base_classifier_require_dense",
            "labelset_size",
        ]

    def fit(self, X, y):
        """Fit classifier to multi-label data

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse
            input features, can be a dense or sparse matrix of size
            :code:`(n_samples, n_features)`
        y : numpy.ndaarray or scipy.sparse {0,1}
            binary indicator matrix with label assignments, shape
            :code:`(n_samples, n_labels)`

        Returns
        -------
        fitted instance of self
        """
        self._label_count = y.shape[1]
        self.model_count_ = int(np.ceil(self._label_count / self.labelset_size))
        self.classifier_ = LabelSpacePartitioningClassifier(
            classifier=LabelPowerset(
                classifier=self.base_classifier,
                require_dense=self.base_classifier_require_dense,
            ),
            clusterer=RandomLabelSpaceClusterer(
                cluster_size=self.labelset_size,
                cluster_count=self.model_count_,
                allow_overlap=False,
            ),
            require_dense=[False, False],
        )
        return self.classifier_.fit(X, y)

    def predict(self, X):
        """Predict label assignments

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of int
            binary indicator matrix with label assignments with shape
            :code:`(n_samples, n_labels)`
        """

        return self.classifier_.predict(X)

    def predict_proba(self, X):
        """Predict label probabilities

        Parameters
        ----------
        X : numpy.ndarray or scipy.sparse.csc_matrix
            input features of shape :code:`(n_samples, n_features)`

        Returns
        -------
        scipy.sparse of float
            binary indicator matrix with probability of label assignment with shape
            :code:`(n_samples, n_labels)`
        """

        return self.classifier_.predict_proba(X)



"""
Weighted Multi-dimensional Scaling (MDS)
"""

# author: Nelle Varoquaux <nelle.varoquaux@gmail.com>
# Licence: BSD
# from: https://raw.githubusercontent.com/ej0cl6/csmlc/master/models/mdsw.py

import numpy as np
import warnings

from sklearn.base import BaseEstimator
from sklearn.metrics import euclidean_distances
from sklearn.utils import check_random_state, check_array, check_symmetric
from joblib import Parallel
from joblib import delayed
from sklearn.isotonic import IsotonicRegression


def _smacof_single_w(
    similarities,
    n_uq,
    uq_weight,
    metric=True,
    n_components=2,
    init=None,
    max_iter=300,
    verbose=0,
    eps=1e-3,
    random_state=None,
):
    """
    Computes multidimensional scaling using SMACOF algorithm

    Parameters
    ----------
    similarities: symmetric ndarray, shape [n * n]
        similarities between the points

    metric: boolean, optional, default: True
        compute metric or nonmetric SMACOF algorithm

    n_components: int, optional, default: 2
        number of dimension in which to immerse the similarities
        overwritten if initial array is provided.

    init: {None or ndarray}, optional
        if None, randomly chooses the initial configuration
        if ndarray, initialize the SMACOF algorithm with this array

    max_iter: int, optional, default: 300
        Maximum number of iterations of the SMACOF algorithm for a single run

    verbose: int, optional, default: 0
        level of verbosity

    eps: float, optional, default: 1e-6
        relative tolerance w.r.t stress to declare converge

    random_state: integer or numpy.RandomState, optional
        The generator used to initialize the centers. If an integer is
        given, it fixes the seed. Defaults to the global numpy random
        number generator.

    Returns
    -------
    X: ndarray (n_samples, n_components), float
               coordinates of the n_samples points in a n_components-space

    stress_: float
        The final value of the stress (sum of squared distance of the
        disparities and the distances for all constrained points)

    n_iter : int
        Number of iterations run.

    """
    similarities = check_symmetric(similarities, raise_exception=True)

    n_samples = similarities.shape[0]
    random_state = check_random_state(random_state)

    W = np.ones((n_samples, n_samples))
    W[:n_uq, :n_uq] = 0.0
    W[n_uq:, n_uq:] = 0.0
    # W[np.arange(len(W)), np.arange(len(W))] = 0.0

    if uq_weight is not None:
        W[:n_uq, n_uq:] *= uq_weight.reshape((uq_weight.shape[0], -1))
        W[n_uq:, :n_uq] *= uq_weight.reshape((-1, uq_weight.shape[0]))

    V = -W
    V[np.arange(len(V)), np.arange(len(V))] = W.sum(axis=1)
    e = np.ones((n_samples, 1))

    Vp = np.linalg.inv(V + np.dot(e, e.T) / n_samples) - np.dot(e, e.T) / n_samples
    # Vp = np.linalg.pinv(V)

    sim_flat = ((1 - np.tri(n_samples)) * similarities).ravel()
    sim_flat_w = sim_flat[sim_flat != 0]
    if init is None:
        # Randomly choose initial configuration
        X = random_state.rand(n_samples * n_components)
        X = X.reshape((n_samples, n_components))
    else:
        # overrides the parameter p
        n_components = init.shape[1]
        if n_samples != init.shape[0]:
            raise ValueError(
                "init matrix should be of shape (%d, %d)" % (n_samples, n_components)
            )
        X = init

    old_stress = None
    ir = IsotonicRegression()
    for it in range(max_iter):
        # Compute distance and monotonic regression
        dis = euclidean_distances(X)

        if metric:
            disparities = similarities
        else:
            dis_flat = dis.ravel()
            # similarities with 0 are considered as missing values
            dis_flat_w = dis_flat[sim_flat != 0]

            # Compute the disparities using a monotonic regression
            disparities_flat = ir.fit_transform(sim_flat_w, dis_flat_w)
            disparities = dis_flat.copy()
            disparities[sim_flat != 0] = disparities_flat
            disparities = disparities.reshape((n_samples, n_samples))
            disparities *= np.sqrt(
                (n_samples * (n_samples - 1) / 2) / (disparities**2).sum()
            )

        # Compute stress
        # stress = ((dis.ravel() - disparities.ravel()) ** 2).sum() / 2
        _stress = (W.ravel() * ((dis.ravel() - disparities.ravel()) ** 2)).sum() / 2

        # Update X using the Guttman transform
        # dis[dis == 0] = 1e-5
        # ratio = disparities / dis
        # B = - ratio
        # B[np.arange(len(B)), np.arange(len(B))] += ratio.sum(axis=1)
        # X = 1. / n_samples * np.dot(B, X)
        # print (1. / n_samples * np.dot(B, X))[:5].T

        dis[dis == 0] = 1e-5
        ratio = disparities / dis
        _B = -W * ratio
        _B[np.arange(len(_B)), np.arange(len(_B))] += (W * ratio).sum(axis=1)

        X = np.dot(Vp, np.dot(_B, X))
        # print X[:5].T

        dis = np.sqrt((X**2).sum(axis=1)).sum()

        if verbose >= 2:
            print("it: %d, stress %s" % (it, _stress))
        if old_stress is not None:
            if (old_stress - _stress / dis) < eps:
                if verbose:
                    print("breaking at iteration %d with stress %s" % (it, _stress))
                break
        old_stress = _stress / dis

    return X, _stress, it + 1


def _smacof_w(
    similarities,
    n_uq,
    uq_weight,
    metric=True,
    n_components=2,
    init=None,
    n_init=8,
    n_jobs=1,
    max_iter=300,
    verbose=0,
    eps=1e-3,
    random_state=None,
    return_n_iter=False,
):
    """
    Computes multidimensional scaling using SMACOF (Scaling by Majorizing a
    Complicated Function) algorithm

    The SMACOF algorithm is a multidimensional scaling algorithm: it minimizes
    a objective function, the *stress*, using a majorization technique. The
    Stress Majorization, also known as the Guttman Transform, guarantees a
    monotone convergence of Stress, and is more powerful than traditional
    techniques such as gradient descent.

    The SMACOF algorithm for metric MDS can summarized by the following steps:

    1. Set an initial start configuration, randomly or not.
    2. Compute the stress
    3. Compute the Guttman Transform
    4. Iterate 2 and 3 until convergence.

    The nonmetric algorithm adds a monotonic regression steps before computing
    the stress.

    Parameters
    ----------
    similarities : symmetric ndarray, shape (n_samples, n_samples)
        similarities between the points

    metric : boolean, optional, default: True
        compute metric or nonmetric SMACOF algorithm

    n_components : int, optional, default: 2
        number of dimension in which to immerse the similarities
        overridden if initial array is provided.

    init : {None or ndarray of shape (n_samples, n_components)}, optional
        if None, randomly chooses the initial configuration
        if ndarray, initialize the SMACOF algorithm with this array

    n_init : int, optional, default: 8
        Number of time the smacof_p algorithm will be run with different
        initialisation. The final results will be the best output of the
        n_init consecutive runs in terms of stress.

    n_jobs : int, optional, default: 1

        The number of jobs to use for the computation. This works by breaking
        down the pairwise matrix into n_jobs even slices and computing them in
        parallel.

        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging. For n_jobs below -1,
        (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all CPUs but one
        are used.

    max_iter : int, optional, default: 300
        Maximum number of iterations of the SMACOF algorithm for a single run

    verbose : int, optional, default: 0
        level of verbosity

    eps : float, optional, default: 1e-6
        relative tolerance w.r.t stress to declare converge

    random_state : integer or numpy.RandomState, optional
        The generator used to initialize the centers. If an integer is
        given, it fixes the seed. Defaults to the global numpy random
        number generator.

    return_n_iter : bool
        Whether or not to return the number of iterations.

    Returns
    -------
    X : ndarray (n_samples,n_components)
        Coordinates of the n_samples points in a n_components-space

    stress : float
        The final value of the stress (sum of squared distance of the
        disparities and the distances for all constrained points)

    n_iter : int
        The number of iterations corresponding to the best stress.
        Returned only if `return_n_iter` is set to True.

    Notes
    -----
    "Modern Multidimensional Scaling - Theory and Applications" Borg, I.;
    Groenen P. Springer Series in Statistics (1997)

    "Nonmetric multidimensional scaling: a numerical method" Kruskal, J.
    Psychometrika, 29 (1964)

    "Multidimensional scaling by optimizing goodness of fit to a nonmetric
    hypothesis" Kruskal, J. Psychometrika, 29, (1964)
    """

    similarities = check_array(similarities)
    random_state = check_random_state(random_state)

    if hasattr(init, "__array__"):
        init = np.asarray(init).copy()
        if not n_init == 1:
            warnings.warn(
                "Explicit initial positions passed: "
                "performing only one init of the MDS instead of %d" % n_init
            )
            n_init = 1

    best_pos, best_stress = None, None

    if n_jobs == 1:
        for it in range(n_init):
            pos, stress, n_iter_ = _smacof_single_w(
                similarities,
                n_uq,
                uq_weight,
                metric=metric,
                n_components=n_components,
                init=init,
                max_iter=max_iter,
                verbose=verbose,
                eps=eps,
                random_state=random_state,
            )
            if best_stress is None or stress < best_stress:
                best_stress = stress
                best_pos = pos.copy()
                best_iter = n_iter_
    else:
        seeds = random_state.randint(np.iinfo(np.int32).max, size=n_init)
        results = Parallel(n_jobs=n_jobs, verbose=max(verbose - 1, 0))(
            delayed(_smacof_single_w)(
                similarities,
                n_uq,
                uq_weight,
                metric=metric,
                n_components=n_components,
                init=init,
                max_iter=max_iter,
                verbose=verbose,
                eps=eps,
                random_state=seed,
            )
            for seed in seeds
        )
        positions, stress, n_iters = zip(*results)
        best = np.argmin(stress)
        best_stress = stress[best]
        best_pos = positions[best]
        best_iter = n_iters[best]

    if return_n_iter:
        return best_pos, best_stress, best_iter
    else:
        return best_pos, best_stress


class _MDSW(BaseEstimator):
    """Multidimensional scaling

    Parameters
    ----------
    metric : boolean, optional, default: True
        compute metric or nonmetric SMACOF (Scaling by Majorizing a
        Complicated Function) algorithm

    n_components : int, optional, default: 2
        number of dimension in which to immerse the similarities
        overridden if initial array is provided.

    n_init : int, optional, default: 4
        Number of time the smacof_p algorithm will be run with different
        initialisation. The final results will be the best output of the
        n_init consecutive runs in terms of stress.

    max_iter : int, optional, default: 300
        Maximum number of iterations of the SMACOF algorithm for a single run

    verbose : int, optional, default: 0
        level of verbosity

    eps : float, optional, default: 1e-6
        relative tolerance w.r.t stress to declare converge

    n_jobs : int, optional, default: 1
        The number of jobs to use for the computation. This works by breaking
        down the pairwise matrix into n_jobs even slices and computing them in
        parallel.

        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging. For n_jobs below -1,
        (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all CPUs but one
        are used.

    random_state : integer or numpy.RandomState, optional
        The generator used to initialize the centers. If an integer is
        given, it fixes the seed. Defaults to the global numpy random
        number generator.

    dissimilarity : string
        Which dissimilarity measure to use.
        Supported are 'euclidean' and 'precomputed'.


    Attributes
    ----------
    embedding_ : array-like, shape [n_components, n_samples]
        Stores the position of the dataset in the embedding space

    stress_ : float
        The final value of the stress (sum of squared distance of the
        disparities and the distances for all constrained points)


    References
    ----------
    "Modern Multidimensional Scaling - Theory and Applications" Borg, I.;
    Groenen P. Springer Series in Statistics (1997)

    "Nonmetric multidimensional scaling: a numerical method" Kruskal, J.
    Psychometrika, 29 (1964)

    "Multidimensional scaling by optimizing goodness of fit to a nonmetric
    hypothesis" Kruskal, J. Psychometrika, 29, (1964)

    """

    def __init__(
        self,
        n_components=2,
        n_uq=1,
        uq_weight=None,
        metric=True,
        n_init=4,
        max_iter=300,
        verbose=0,
        eps=1e-3,
        n_jobs=1,
        random_state=None,
        dissimilarity="euclidean",
    ):
        self.n_components = n_components
        self.n_uq = n_uq
        self.uq_weight = uq_weight
        self.dissimilarity = dissimilarity
        self.metric = metric
        self.n_init = n_init
        self.max_iter = max_iter
        self.eps = eps
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.random_state = random_state

    @property
    def _pairwise(self):
        return self.kernel == "precomputed"

    def fit(self, X, y=None, init=None):
        """
        Computes the position of the points in the embedding space

        Parameters
        ----------
        X : array, shape=[n_samples, n_features], or [n_samples, n_samples] \
                if dissimilarity='precomputed'
            Input data.

        init : {None or ndarray, shape (n_samples,)}, optional
            If None, randomly chooses the initial configuration
            if ndarray, initialize the SMACOF algorithm with this array.
        """
        self.fit_transform(X, init=init)
        return self

    def fit_transform(self, X, y=None, init=None):
        """
        Fit the data from X, and returns the embedded coordinates

        Parameters
        ----------
        X : array, shape=[n_samples, n_features], or [n_samples, n_samples] \
                if dissimilarity='precomputed'
            Input data.

        init : {None or ndarray, shape (n_samples,)}, optional
            If None, randomly chooses the initial configuration
            if ndarray, initialize the SMACOF algorithm with this array.

        """
        X = check_array(X)
        if X.shape[0] == X.shape[1] and self.dissimilarity != "precomputed":
            warnings.warn(
                "The MDS API has changed. ``fit`` now constructs an"
                " dissimilarity matrix from data. To use a custom "
                "dissimilarity matrix, set "
                "``dissimilarity=precomputed``."
            )

        if self.dissimilarity == "precomputed":
            self.dissimilarity_matrix_ = X
        elif self.dissimilarity == "euclidean":
            self.dissimilarity_matrix_ = euclidean_distances(X)
        else:
            raise ValueError(
                "Proximity must be 'precomputed' or 'euclidean'."
                " Got %s instead" % str(self.dissimilarity)
            )

        self.embedding_, self.stress_, self.n_iter_ = _smacof_w(
            self.dissimilarity_matrix_,
            self.n_uq,
            self.uq_weight,
            metric=self.metric,
            n_components=self.n_components,
            init=init,
            n_init=self.n_init,
            n_jobs=self.n_jobs,
            max_iter=self.max_iter,
            verbose=self.verbose,
            eps=self.eps,
            random_state=self.random_state,
            return_n_iter=True,
        )

        return self.embedding_
